//-----------------------------------------------------------------------------
// Name: DIGame.cpp
//
// Desc: Demonstrates a little game which uses DirectInput. This file contains
//       all the DInput code. The other file contains all the game play code.
//
// Copyright (c) 1998-1999 Microsoft Corp.  All rights reserved.
// Ported by Vadim Bodrov, TMT Development Corp.
//-----------------------------------------------------------------------------
uses Windows, Messages, MMSystem, DXLite, GamePlay;

{$i resource.inc}
{$r digame.res}

//-----------------------------------------------------------------------------
// Global variables and functions
//-----------------------------------------------------------------------------
const
    NUM_ROWS       = 8;
    WINDOW_WIDTH   = 320;
    WINDOW_HEIGHT  = (NUM_ROWS+1)*40+25;
    WM_GAMEOVER    = WM_USER+1;
    WM_SYNCACQUIRE = WM_USER+2;

var
    g_pDI:               PPDirectInput7;        // The DInput object
    g_pdidDevice2:       PPDirectInputDevice2;  // The DIDevice2 interface
    g_guidJoystick:      TGUID;                 // The GUID for the joystick
    g_bActive:           Boolean;               // Whether app is active
    g_bGamePlaying:      Boolean;               // Whether app is playing

    g_bUseKeyboard:      Boolean := TRUE;
    g_bUseMouse:         Boolean := FALSE;
    g_bUseJoystick:      Boolean := FALSE;

//-----------------------------------------------------------------------------
// Name: DisplayError()
// Desc: Displays an error message.
//-----------------------------------------------------------------------------
procedure DisplayError( strMessage: PChar );
begin
    MessageBox( 0, strMessage, 'DInput Sample Game', MB_OK );
end;




//-----------------------------------------------------------------------------
// Name: EnumJoysticksCallback()
// Desc: Called once for each enumerated joystick. If we find one,
//       create a device interface on it so we can play with it.
//-----------------------------------------------------------------------------
function EnumJoysticksCallback conv arg_stdcall(pInst: LPDIDEVICEINSTANCEA;
    pvContext: Pointer ): LongBool;
begin
    CopyMemory( pvContext, @pInst^.guidInstance, sizeof(TGUID) );

    Result := DIENUM_STOP;
end;




//-----------------------------------------------------------------------------
// Name: CreateDInput()
// Desc: Initialize the DirectInput objects.
//-----------------------------------------------------------------------------
function CreateDInput( Wnd: HWND ): HResult;
begin
    // Create the main DirectInput object
    if( FAILED( DirectInputCreateEx( GetWindowLong( Wnd, GWL_HINSTANCE ),
                                   DIRECTINPUT_VERSION, IID_IDirectInput7, @g_pDI, nil) ) ) then

    begin
        DisplayError( 'DirectInputCreate() failed.' );
        Result := E_FAIL;
        exit;
    end;

    // Check to see if a joystick is present. If so, the enumeration callback
    // will save the joystick's GUID, so we can create it later.
    ZeroMemory( @g_guidJoystick, sizeof(TGUID) );

    g_pDI^^.EnumDevices^(g_pDI, DIDEVTYPE_JOYSTICK, TDIEnumDevicesCallbackA(@EnumJoysticksCallback),
                        @g_guidJoystick, DIEDFL_ATTACHEDONLY );

    Result := S_OK;
end;


//-----------------------------------------------------------------------------
// Name: CreateInputDevice()
// Desc: Create a DirectInput device.
//-----------------------------------------------------------------------------
function CreateInputDevice( Wnd: HWND; guidDevice: TGUID;
                           const pdidDataFormat: TDIDATAFORMAT; dwFlags: DWORD ): HResult;
begin
    // Obtain an interface to the input device
    if( FAILED( g_pDI^^.CreateDeviceEx^(g_pDI, guidDevice, IID_IDirectInputDevice2,
        @g_pdidDevice2, nil ) ) ) then
    begin
        DisplayError( 'CreateDeviceEx() failed' );
        Result := E_FAIL;
        exit;
    end;


    // Set the device data format. Note: a data format specifies which
    // controls on a device we are interested in, and how they should be
    // reported.
    if( FAILED( g_pdidDevice2^^.SetDataFormat^(g_pdidDevice2, pdidDataFormat ) ) ) then
    begin
        DisplayError( 'SetDataFormat() failed' );
        Result := E_FAIL;
        exit;
    end;

    // Set the cooperativity level to let DirectInput know how this device
    // should interact with the system and with other DirectInput applications.
    if( FAILED( g_pdidDevice2^^.SetCooperativeLevel^(g_pdidDevice2, Wnd, dwFlags ) ) ) then
    begin
        DisplayError( 'SetCooperativeLevel() failed' );
        Result := E_FAIL;
        exit;
    end;

    Result := S_OK;
end;




//-----------------------------------------------------------------------------
// Name: DestroyInputDevice()
// Desc: Release the DirectInput device
//-----------------------------------------------------------------------------
procedure DestroyInputDevice;
begin
    // Unacquire and release the device's interfaces
    if g_pdidDevice2 <> nil then
    begin
        g_pdidDevice2^^.Unacquire^(g_pdidDevice2);
        g_pdidDevice2^^.Release^(g_pdidDevice2);
        g_pdidDevice2 := nil;
    end;
end;




//-----------------------------------------------------------------------------
// Name: DestroyDInput()
// Desc: Terminate our usage of DirectInput
//-----------------------------------------------------------------------------
procedure DestroyDInput;
begin
    // Destroy the DInput object
    if g_pDI <> nil then
        g_pDI^^.Release^(g_pDI);
    g_pDI := nil;
end;



var
    fUpTime: Double;
    fDnTime: Double;


//-----------------------------------------------------------------------------
// Name: UpdateFrame()
// Desc: The game plays here. Use DInput to get input. Move the player's ship
//       accordingly, update the weapons' positions and handle collisions.
//-----------------------------------------------------------------------------
procedure UpdateFrame( Wnd: HWND );
begin
    g_fCurrentTime := timeGetTime * 0.001;

    if g_pdidDevice2 <> nil then
    declare
    var
        hr: HREsult;
        diks: array[0..255] of Byte; // DInput keyboard state buffer
        dims: TDIMouseState;         // DInput mouse state structure
        dijs: TDIJoyState;           // DInput joystick state structure
    begin
        // Poll the device before reading the current state. This is required
        // for some devices (joysticks) but has no effect for others (keyboard
        // and mice). Note: this uses a DIDevice2 interface for the device.
        if FAILED( g_pdidDevice2^^.Poll^(g_pdidDevice2)) then
            exit;

        if g_bUseKeyboard then // Get the keyboard state
            hr := g_pdidDevice2^^.GetDeviceState^(g_pdidDevice2, sizeof(diks), diks );

        if g_bUseMouse then    // Else, get the mouse state
            hr := g_pdidDevice2^^.GetDeviceState^(g_pdidDevice2, sizeof(DIMOUSESTATE), dims );

        if g_bUseJoystick then // Else, get the joystick state
            hr := g_pdidDevice2^^.GetDeviceState^(g_pdidDevice2, sizeof(DIJOYSTATE), dijs );

        // Check whether the input stream has been interrupted. If so,
        // re-acquire and try again.
        if hr = DIERR_INPUTLOST then
        begin
            hr := g_pdidDevice2^^.Acquire^(g_pdidDevice2);
            if SUCCEEDED(hr) then
                exit;
        end;

        // Read keyboard input
        if g_bUseKeyboard then
        begin
            // Update the variables for the player's ship
            MovePlayer( (diks[DIK_LEFT] and $80) <> 0, (diks[DIK_RIGHT] and $80) <> 0,
                        (diks[DIK_UP] and $80) <> 0, (diks[DIK_DOWN] and $80) <> 0,
                        (diks[DIK_SPACE] and $80) <> 0 );
        end;

        // Read mouse input
        if g_bUseMouse then
        begin
            // Note: on really fast computers, the mouse will appear to be
            // still most of the time, and move in jumps. To combat this, we'll
            // keep 0.1 seconds of persistence for any up/down values we read.
            fUpTime := 0.0;
            fDnTime := 0.0;
            if dims.lY < 0  then begin
                fDnTime := 0.0;
                fUpTime := g_fCurrentTime+0.1;
            end;
            if dims.lY > 0 then
            begin
                fUpTime := 0.0;
                fDnTime := g_fCurrentTime+0.1;
            end;
            MovePlayer( dims.lX < 0, dims.lX > 0, (fUpTime - g_fCurrentTime) > 0.0,
                        (fDnTime - g_fCurrentTime) > 0.0, (dims.rgbButtons[0] and $80) <> 0 );
        end;

        // Read joystick input
        if g_bUseJoystick then
        begin
            // Update the variables for the player's ship
            MovePlayer( dijs.lX < 0, dijs.lX > 0, dijs.lY < 0, dijs.lY > 0,
                        (dijs.rgbButtons[0] and $80) <> 0 );
        end;
    end;

    AnimateScene( Wnd );
    RenderScene( Wnd );
end;




//-----------------------------------------------------------------------------
// Name: InputDeviceSelectProc()
// Desc: Dialog procedure for selecting an input device.
//-----------------------------------------------------------------------------
function InputDeviceSelectProc conv arg_stdcall ( Wnd: HWND; msg: UINT; wParam: WPARAM;
    lParam: LPARAM ): LONGBOOL;
var
    hwndKeyboardButton: HWND := GetDlgItem( Wnd, IDC_KEYBOARD );
    hwndMouseButton: HWND    := GetDlgItem( Wnd, IDC_MOUSE );
    hwndJoystickButton: HWND := GetDlgItem( Wnd, IDC_JOYSTICK );
begin
    if WM_INITDIALOG = msg then
    begin
        SendMessage( hwndKeyboardButton, BM_SETCHECK, DWORD(g_bUseKeyboard), 0);
        SendMessage( hwndMouseButton,    BM_SETCHECK, DWORD(g_bUseMouse),    0);
        SendMessage( hwndJoystickButton, BM_SETCHECK, DWORD(g_bUseJoystick), 0);

        EnableWindow( hwndJoystickButton, not isGUID_NULL(g_guidJoystick) );

        Result := LONGTRUE;
        exit;
    end;

    if (WM_COMMAND = msg) and (IDOK = LOWORD(wParam) ) then
    begin
        // Destroy the old device
        DestroyInputDevice;

        // Check the dialog controls to see which device type to create
        g_bUseKeyboard := Boolean(SendMessage( hwndKeyboardButton, BM_GETCHECK, 0, 0));
        g_bUseMouse    := Boolean(SendMessage( hwndMouseButton,    BM_GETCHECK, 0, 0));
        g_bUseJoystick := Boolean(SendMessage( hwndJoystickButton, BM_GETCHECK, 0, 0));

        if g_bUseKeyboard then
        begin
            CreateInputDevice( GetParent(Wnd), GUID_SysKeyboard, c_dfDIKeyboard,
                               DISCL_NONEXCLUSIVE or DISCL_FOREGROUND );
        end
        else if g_bUseMouse then
        begin
            CreateInputDevice( GetParent(Wnd), GUID_SysMouse, c_dfDIMouse,
                               DISCL_EXCLUSIVE or DISCL_FOREGROUND );
        end
        else if g_bUseJoystick then
        begin
            CreateInputDevice( GetParent(Wnd), g_guidJoystick, c_dfDIJoystick,
                               DISCL_EXCLUSIVE or DISCL_FOREGROUND );

            // Set the range of the joystick axes tp [-1000,+1000]
            declare
            var
                diprg: TDIPropRange;
                dipdw: TDIPropDWORD;
            begin
                diprg.diph.dwSize       := sizeof(TDIPropRange);
                diprg.diph.dwHeaderSize := sizeof(TDIPropRange);
                diprg.diph.dwHow        := DIPH_BYOFFSET;
                diprg.lMin              := -10;
                diprg.lMax              := +10;

                diprg.diph.dwObj := DIJOFS_X;    // Set the x-axis range
                g_pdidDevice2^^.SetProperty^(g_pdidDevice2, DIPROP_RANGE, diprg.diph);

                diprg.diph.dwObj := DIJOFS_Y;    // Set the y-axis range
                g_pdidDevice2^^.SetProperty^(g_pdidDevice2, DIPROP_RANGE, diprg.diph);

                // Set the dead zone for the joystick axes (because many joysticks
                // aren't perfectly calibrated to be zero when centered).
                dipdw.diph.dwSize       := sizeof(TDIPropDWORD);
                dipdw.diph.dwHeaderSize := sizeof(TDIPROPHEADER);
                dipdw.diph.dwHow        := DIPH_DEVICE;
                dipdw.dwData            := 1000; // Here, 1000 = 10%

                dipdw.diph.dwObj := DIJOFS_X;    // Set the x-axis deadzone
                g_pdidDevice2^^.SetProperty^(g_pdidDevice2, DIPROP_DEADZONE, dipdw.diph);

                dipdw.diph.dwObj := DIJOFS_Y;    // Set the y-axis deadzone
                g_pdidDevice2^^.SetProperty^(g_pdidDevice2, DIPROP_RANGE, dipdw.diph);
            end;
        end;

        EndDialog( Wnd, IDOK );
        Result := LONGTRUE;
        exit;
    end;

    if (WM_COMMAND = msg) and (IDCANCEL = LOWORD(wParam)) then
    begin
        EndDialog( Wnd, IDCANCEL );
        Result := LONGTRUE;
        exit;
    end;

    Result := LONGFALSE;
end;



//-----------------------------------------------------------------------------
// Name: WndProc()
// Desc: Window procedure for the game window
//-----------------------------------------------------------------------------
function WndProc conv arg_stdcall ( Wnd: HWND; msg: UINT; wParam: WPARAM; lParam: LPARAM ): LResult;
begin
    case msg of
        WM_ACTIVATE:
        begin
            // Pause or unpause the app (and acquire or unacquire the device)
            // based on whether we are losing or gaining activation.
            g_bActive :=  WA_INACTIVE <> wParam ;
            PauseGame( not g_bActive );
            PostMessage( Wnd, WM_SYNCACQUIRE, 0, 0 );
        end;

        WM_ENTERMENULOOP:
        begin
            g_bActive := FALSE;
            PauseGame( TRUE );
            PostMessage( Wnd, WM_SYNCACQUIRE, 0, 0 );
        end;

        WM_EXITMENULOOP:
        begin
            if ( GetActiveWindow <> Wnd ) and ( not IsIconic( Wnd ) ) then
                g_bActive := TRUE;
            PauseGame( FALSE );
            PostMessage( Wnd, WM_SYNCACQUIRE, 0, 0 );
        end;

        WM_PAINT:
        begin
            RenderScene( Wnd );
        end;

        WM_SYNCACQUIRE:
        begin
            if g_pdidDevice2 <> nil then
            begin
                if g_bActive and g_bGamePlaying then
                    g_pdidDevice2^^.Acquire^(g_pdidDevice2)
                else
                    g_pdidDevice2^^.Unacquire^(g_pdidDevice2);
            end;
        end;

        WM_GAMEOVER:
        begin
            g_bGamePlaying := FALSE;
            PostMessage( Wnd, WM_SYNCACQUIRE, 0, 0 );
        end;

        WM_COMMAND:
        begin
            case LOWORD(wParam) of
                IDM_NEWGAME:
                begin
                    StartNewGame;
                    g_bGamePlaying := TRUE;
                    g_bActive := TRUE;
                    PostMessage( Wnd, WM_SYNCACQUIRE, 0, 0 );
                end;

                IDM_SELECTINPUTDEVICE:
                begin
                    DialogBox( GetWindowLong( Wnd, GWL_HINSTANCE ),
                               MAKEINTRESOURCE(IDD_SELECTINPUTDEVICE), Wnd,
                               @InputDeviceSelectProc );
                end;

                IDM_EXIT:
                begin
                    SendMessage( Wnd, WM_CLOSE, 0, 0 );
                    Result := 0;
                    exit;
                end;
            end;
        end;

        WM_DESTROY:
        begin
            PostQuitMessage(0);
        end;
    end;

    Result := DefWindowProc( Wnd, msg, wParam, lParam );
end;




//-----------------------------------------------------------------------------
// WinMain()
// Desc: Application entry point.
//-----------------------------------------------------------------------------
var
    wc: TWndClass;
    Wnd: HWND;
    msg: TMsg;
begin
    // Set up the window class

    wc.style := CS_HREDRAW or CS_VREDRAW;
    wc.lpfnWndProc := @WndProc;
    wc.cbClsExtra := 0;
    wc.cbWndExtra := 0;
    wc.hInstance := hInstance;
    wc.hIcon := LoadIcon( hInstance, MAKEINTRESOURCE(IDI_DIRECTX_ICON));
    wc.hCursor := LoadCursor(0, IDC_ARROW);
    wc.hbrBackground := 0;
    wc.lpszMenuName :=  MAKEINTRESOURCE(IDR_MENU);
    wc.lpszClassName := 'DI Window';

    if RegisterClass( wc ) = 0 then
        halt(0);

    // Create the main window
    Wnd := CreateWindow( 'DI Window', 'GameInput - Simple Game Using DInput',
                         WS_OVERLAPPED or WS_CAPTION or WS_SYSMENU or WS_MINIMIZEBOX,
                         CW_USEDEFAULT, CW_USEDEFAULT,
                         WINDOW_WIDTH, WINDOW_HEIGHT, 0, 0, hInstance, nil );

    // Create the DInput object
    if FAILED( CreateDInput( Wnd ) ) then
    begin
        DestroyWindow( Wnd );
        halt(0);
    end;

    // Create a keyboard device
    if FAILED( CreateInputDevice( Wnd, GUID_SysKeyboard, c_dfDIKeyboard,
                                 DISCL_NONEXCLUSIVE or DISCL_FOREGROUND ) ) then
    begin
        DestroyWindow( Wnd );
        halt(0);
    end;

    // Initialize and start a new game
    ShowWindow( Wnd, CmdShow );
    InitializeGameObjects( hInstance, Wnd );
    StartNewGame;
    g_bGamePlaying := TRUE;

    // Standard game loop.
    while TRUE do
    begin
        if PeekMessage( msg, 0, 0, 0, PM_REMOVE ) then
        begin
            // Check for a quit message
            if msg.message = WM_QUIT then
                break;

            TranslateMessage( msg );
            DispatchMessage( msg );
        end
        else if (g_bActive and g_bGamePlaying) then
        begin
            // Play the game (check user input and update the window)
            UpdateFrame( Wnd );
        end
        else
            WaitMessage;
    end;

    // Cleanup before exitting
    DeleteGameObjects;
    DestroyInputDevice;
    DestroyDInput;
end.
